from typing import Callable, Any, List
from .JPTestBook import JPTestBook


class JPPreRun:
    """
    decorator to use with functions that should be executed prior to all tests
    """
    FN: List[Callable[[], Any]] = []

    def __init__(self, f: Callable[[], Any]):
        self.FN.append(f)


class JPPreTest:
    """
    decorator to use with functions that should be executed prior to any test
    """
    FN: List[Callable[[JPTestBook], Any]] = []

    def __init__(self, f: Callable[[JPTestBook], Any]):
        self.FN.append(f)


class JPPostTest:
    """
    decorator to use with functions that should be executed after any test
    """
    FN: List[Callable[[JPTestBook], Any]] = []

    def __init__(self, f: Callable[[JPTestBook], Any]):
        self.FN.append(f)


class JPPostRun:
    """
    decorator to use with functions that should be executed after all tests
    """
    FN: List[Callable[[], Any]] = []

    def __init__(self, f: Callable[[], Any]):
        self.FN.append(f)



import importlib.util
import json
import sys
from argparse import ArgumentParser
from typing import Dict, Tuple, List, Optional

from jptest import *

# configure argument parser
parser = ArgumentParser()
parser.add_argument('nb_file', help='notebook file (.ipynb) to load')
parser.add_argument('test_file', help='test file (.py) to load', nargs='?')
parser.add_argument('test_name', help='test to execute (all if None given)', nargs='?')
parser.add_argument('--json', action='store_true', help='print output as json (default)', default=True)
parser.add_argument('--md', action='store_true', help='print output as markdown')
parser.add_argument('--quiet', action='store_true', help='only print exceptions')
parser.add_argument('--verbose', '-v', action='store_true', help='print verbosely to stderr')

args = parser.parse_args()

# load classes from file
JPTest.NOTEBOOK = args.nb_file

if args.test_file is not None:
    spec = importlib.util.spec_from_file_location('testfile', args.test_file)
    foo = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = foo
    spec.loader.exec_module(foo)
else:
    # noinspection PyUnresolvedReferences
    from jptest.RunTest import *

# pre run functions
if args.verbose:
    print('pre run', file=sys.stderr)

for fun in JPPreRun.FN:
    fun()

# execute tests
achieved, total = 0.0, 0.0
tests: Dict[str, Tuple[str, float, float, List[str], Optional[Exception]]] = {}

for test, original_fun, wrapped_fun in JPTest.TESTS:
    test_name = original_fun.__name__

    if args.verbose:
        print(test_name, file=sys.stderr)

    # skip tests if required
    if args.test_name is not None and args.test_name != test_name:
        continue

    # execute test
    try:
        score, comments = wrapped_fun()

        achieved += score
        tests[test_name] = test.name, score, test.max_score, comments, None
    except Exception as e:
        tests[test_name] = test.name, 0, test.max_score, [str(e)], e

    total += test.max_score

# post run functions
if args.verbose:
    print('post run', file=sys.stderr)

for fun in JPPostRun.FN:
    fun()

# print output
if args.quiet:  # quiet
    for test_name, (_, score, max_score, _, e) in tests.items():
        if e is not None:
            raise e

        if score != max_score:
            raise AssertionError(f'test_name={test_name}, score={score}, max_score={max_score}')

elif args.md:  # md
    print(f'# Bewertung ({achieved} / {total})')
    print()

    for test_name, (name, score, max_score, comments, _) in tests.items():
        print(f'## {name} ({score} / {max_score})')
        for comment in comments:
            print(f'- {comment}')
        print()

else:  # json
    result = {
        'achievedScore': achieved,
        'totalScore': total,
        'tests': []
    }

    for test_name, (name, score, max_score, comments, _) in tests.items():
        result['tests'].append({
            'test': test_name,
            'name': name,
            'achievedScore': score,
            'totalScore': max_score,
            'comments': comments
        })

    print(json.dumps(result, indent=4))



import functools
from types import GeneratorType
from typing import List, Union, Tuple, Callable

from testbook import testbook

from .JPFixture import JPPreTest, JPPostTest, JPTestBook
from .JPTestBook import EXECUTE_TYPE


class JPTest(testbook):
    """
    decorator to use with test functions
    """
    NOTEBOOK: str = None
    TESTS: List[Tuple['JPTest', Callable, Callable]] = []

    def __init__(self, name: str = None, max_score: Union[float, int] = None,
                 execute: EXECUTE_TYPE = None, timeout: int = 120,
                 **kwargs):
        """
        :param name: name used in the output
        :param max_score: maximum score (can be exceeded, used to calculate total score)
        :param execute: cells or code to execute prior to the test
        :param timeout: execution timeout in seconds (default: 2 minutes)
        :param kwargs: any named parameters are copied to the JPTestBook object
        """
        super().__init__(self.NOTEBOOK, execute, timeout=timeout)

        self.tb: JPTestBook = JPTestBook(self.client, **kwargs)
        self.name: str = name
        self.max_score: float = None if max_score is None else float(max_score)

    def __call__(self, fun: Callable):
        @functools.wraps(fun)
        def wrapper():
            test_score = 0.0
            test_comments = []

            with self.client.setup_kernel():
                # pre test functions
                for pre_test in JPPreTest.FN:
                    pre_test(self.tb)

                # execution
                if self.execute is None:
                    tracked_parameters = []
                else:
                    tracked_parameters = self.tb.execute(self.execute)

                ret_val = fun(self.tb, *tracked_parameters)
                if isinstance(ret_val, GeneratorType):
                    for value in ret_val:
                        if len(value) == 2:
                            val, score = value
                            pos_comment = None
                            neg_comment = None
                        elif len(value) == 3:
                            val, score, neg_comment = value
                            pos_comment = None
                        elif len(value) == 4:
                            val, score, neg_comment, pos_comment = value
                        else:
                            raise ValueError('invalid yield from test')

                        if (isinstance(val, bool) and val) or (isinstance(val, list) and all(val)):
                            test_score += score
                            if pos_comment is not None:
                                test_comments.append(pos_comment)
                        elif neg_comment is not None:
                            test_comments.append(neg_comment)

                # post test functions
                for post_test in JPPostTest.FN:
                    post_test(self.tb)

            return max(0.0, test_score), test_comments

        wrapper.patchings = [self]

        self.TESTS.append((self, fun, wrapper))
        return wrapper



from typing import List

from testbook.client import TestbookNotebookClient


class JPTestCell:
    @staticmethod
    def from_client(client: TestbookNotebookClient) -> List['JPTestCell']:
        return [JPTestCell(client, idx) for idx in range(len(client.cells))]

    def __init__(self, client: TestbookNotebookClient, idx: int):
        self._client: TestbookNotebookClient = client
        self._idx: int = idx

    @property
    def tags(self) -> List[str]:
        cell = self._client.cells[self._idx]

        if 'tags' not in cell['metadata']:
            return []
        else:
            return cell['metadata']['tags']

    def execute(self):
        self._client.execute_cell(self._idx)



from .JPFixture import JPPreRun, JPPreTest, JPPostTest, JPPostRun
from .JPTestParams import JPTestParams
from .JPTestBook import JPTestBook
from .JPTest import JPTest
from .JPTestCell import JPTestCell



from contextlib import ExitStack
from types import FunctionType
from typing import List, Dict, Union, Callable, Any
from typing import Tuple
from uuid import uuid4

from testbook.client import TestbookNotebookClient

from .JPTestCell import JPTestCell
from .JPTestParams import JPTestParams

TRACK_TYPE = Dict[str,
                  Union[Tuple[str, int],
                        List[Tuple[str, int]]]]
EXECUTE_TYPE = Union[Tuple[str],
                     Tuple[str, str],
                     str,
                     Callable[['JPTestBook'], Any],
                     List['EXECUTE_TYPE'],
                     Dict[str, Union['EXECUTE_TYPE', TRACK_TYPE]]]


class JPTestBook:
    """
    main class for notebook interaction
    """

    def __init__(self, client: TestbookNotebookClient, **kwargs):
        self._client = client

        for key, value in kwargs.items():
            setattr(self, key, value)

    @staticmethod
    def random_id(name: str) -> str:
        random_id = str(uuid4()).replace('-', '_')
        return f'_{name}_{random_id}'

    def get(self, *name: str):
        """
        like `ref`, but copy to a name with a random suffix before

        :param name: object or function name(s)
        :return: reference or a tuple of references depending on parameters
        """
        if len(name) == 1:
            name_with_id = self.random_id(name[0])
            return self.inject(f'{name_with_id} = {name[0]}', name_with_id)
        else:
            return tuple(self.get(n) for n in name)

    def ref(self, *name):
        """
        get a reference to an object or a function

        :param name: object or function name(s)
        :return: reference or a tuple of references depending on parameters
        """
        if len(name) == 1:
            return self._client.ref(name[0])
        else:
            return tuple(self.ref(n) for n in name)

    def inject(self, code: str, *name: str):
        """
        inject code into notebook and return a reference

        :param code: code to inject
        :param name: object or function name(s)
        :return: reference or a tuple of references depending on parameters
        """
        self._client.inject(code)
        return self.ref(*name)

    def track(self, fun_name: str, *parameters: Tuple[str, int]) -> JPTestParams:
        """
        wrap a function to track calls

        :param fun_name: function name
        :param parameters: list of parameters to track
        :return: instance of JPTestParams
        """
        return JPTestParams(self, fun_name, *parameters)

    def cells(self, *item: str) -> List[JPTestCell]:
        cells = JPTestCell.from_client(self._client)

        # no items: all cells
        if len(item) == 0:
            return cells

        # one item: every cell with this tag
        elif len(item) == 1:
            return [cell for cell in cells if item[0] in cell.tags]

        # two items: every cell from the first occurrence of the first tag to
        # the first occurrence of the second tag
        if len(item) == 2:
            result = []

            start = False
            for cell in cells:
                if item[1] in cell.tags:
                    break

                if item[0] in cell.tags:
                    start = True
                if start:
                    result.append(cell)

            return result

        else:
            raise ValueError('unsupported tuple length')

    def execute(self, item: EXECUTE_TYPE) -> List[JPTestParams]:
        """
        :param item: item to execute
        :return: a list of JPTestParams if any track directives were used
        """
        extracted: List[JPTestParams] = []

        # dictionaries
        if isinstance(item, dict):
            with ExitStack() as stack:
                # create function wrappers
                if 'track' in item:
                    for key in item['track']:
                        if isinstance(item['track'][key], list):
                            cm = self.track(key, *item['track'][key])
                        else:
                            cm = self.track(key, item['track'][key])

                        stack.enter_context(cm)
                        extracted.append(cm)

                # execute cells
                for ep in self.execute(item['execute']):
                    extracted.append(ep)

        # tuples
        elif isinstance(item, tuple):
            for cell in self.cells(*item):
                cell.execute()

        # string
        elif isinstance(item, str):
            self._client.inject(item)

        # function
        elif isinstance(item, FunctionType):
            item(self)

        # list
        elif isinstance(item, list):
            for i in item:
                for ep in self.execute(i):
                    extracted.append(ep)

        else:
            raise ValueError(f'unsupported parameter type {type(item)}')

        return extracted



from . import *


@JPTest('run', max_score=0)
def run(tb: JPTestBook):
    # iterate over cells
    for cell in tb.cells():
        # find `jp:input` tag
        replace_input = None

        for tag in cell.tags:
            if tag.startswith('jp:input='):
                replace_input = tag[9:]

        # replace input function if necessary
        replace_name = None

        if replace_input is not None:
            replace_name = JPTestBook.random_id('input')
            tb.inject(f'''
                {replace_name} = input
                input = lambda _: '{replace_input}'
            ''')

        # execute cell
        cell.execute()

        # revert replacement
        if replace_input is not None:
            tb.inject(f'''
                input = {replace_name}
            ''')



from typing import List, Dict, Any, Tuple

from testbook.reference import TestbookObjectReference

from . import JPTestBook


class JPTestParams:
    """
    stores tracked parameters after __exit__ is called
    """
    def __init__(self, tb: JPTestBook, fun_name: str, *parameters: Tuple[str, int]):
        """
        :param tb:
        :param fun_name: function name to replace
        :param parameters: parameters to track as tuples (name, position)
        """
        self._tb = tb
        self._fun_name = fun_name
        self._wrapper: TestbookObjectReference

        self._parameters: Tuple[Tuple[str, int]] = parameters
        self._values: List[Dict[str, Any]] = []

    def __enter__(self):
        """
        replace function with wrapper

        :return: self
        """
        class_name = self._tb.random_id('Track')
        variable_name = self._tb.random_id('track')

        self._wrapper = self._tb.inject(f'''
            class {class_name}:
                def __init__(self, fun):
                    self._calls = []
                    self._results = []

                    def patch(*args, **kwargs):
                        self._calls.append((args, kwargs))
                        result = fun(*args, **kwargs)
                        self._results.append(result)

                        return result

                    self.fun = fun
                    self.patch = patch

                def extract_parameters(self, *parameters):
                    result = []
                    for args, kwargs in self._calls:
                        call_result = {{}}
                        result.append(call_result)

                        for name, pos in parameters:
                            if name in kwargs:
                                call_result[name] = kwargs[name]
                            elif pos < len(args):
                                call_result[name] = args[pos]

                    return result

            {variable_name} = {class_name}({self._fun_name})
            {self._fun_name} = {variable_name}.patch
        ''', variable_name)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        extract parameters and replace wrapper with original function

        :param exc_type:
        :param exc_val:
        :param exc_tb:
        :return:
        """
        self._tb.inject(f'''
            {self._fun_name} = {self._wrapper.name}.fun
        ''')

        self._values = self._wrapper.extract_parameters(*self._parameters)

    @property
    def name(self) -> str:
        return self._fun_name

    @property
    def parameter_names(self) -> List[str]:
        return list(map(lambda x: x[0], self._parameters))

    @property
    def calls(self) -> List[Dict[str, Any]]:
        return self._values

    @property
    def last_call(self) -> Dict[str, Any]:
        return self._values[-1]

    def __len__(self) -> int:
        return len(self.calls)

    def values(self, parameter_name: str) -> List[Any]:
        """
        returns all values for a specific parameter

        :param parameter_name: parameter name
        :return: list of parameter values
        """
        result = []
        for call in self.calls:
            if parameter_name in call:
                result.append(call[parameter_name])

        return result

    def last_value(self, parameter_name: str) -> Any:
        """
        returns the value for a specific parameter in the last call

        :param parameter_name: parameter name
        :return: parameter value
        """
        return self.last_call[parameter_name]
